@@ -16,6 +16,7 @@ import Control.Exception (throw)
 import Control.Monad.Trans.Resource as C (runResourceT)
 import Data.Maybe   (fromMaybe,catMaybes,fromJust)
 import qualified Data.List as DL
+import qualified Data.List.Split as DLS (chunksOf)
 import Control.Monad.ST (runST)
 import Data.Monoid
 import qualified Data.IORef as I
@@ -235,6 +236,12 @@ main = hspec $ do
                     C.=$ CL.fold (+) 0
             x `shouldBe` 2 * sum [1..10 :: Int]
 
+        prop "chunksOf" $ equivToList
+            (DLS.chunksOf 5 :: [Int]->[[Int]]) (CL.chunksOf 5)
+
+        prop "chunksOf (negative)" $ equivToList
+            (map (:[]) :: [Int]->[[Int]]) (CL.chunksOf (-5))
+
         it "groupBy" $ do
             let input = [1::Int, 1, 2, 3, 3, 3, 4, 5, 5]
             x <- runResourceT $ CL.sourceList input
@@ -847,6 +854,15 @@ main = hspec $ do
             x <- I.readIORef ref
             x `shouldBe` (-1)
 
+        it "handles the last input correctly #304" $ do
+            ref <- I.newIORef (-1 :: Int)
+            let sink = CL.mapM_ (I.writeIORef ref)
+                conduit = C.passthroughSink sink (const (return ()))
+            res <- mapM_ C.yield [1..] C.$$ conduit C.=$ CL.take 5
+            res `shouldBe` [1..5]
+            x <- I.readIORef ref
+            x `shouldBe` 5
+
     describe "mtl instances" $ do
         it "ErrorT" $ do
             let src = flip catchError (const $ C.yield 4) $ do
